#include "slideface.h"

#include "slideworld.h"
#include "slidepoint.h"
#include "slidesurface.h"
#include "slidelight.h"

#include "scanconverter.h"

BOOL CSLIDEFace::m_bBackFaceCullingEnabled = TRUE;
BOOL CSLIDEFace::m_bClipEnabled = TRUE;
BOOL CSLIDEFace::m_bRenderNormalsEnabled = FALSE;
FLOAT CSLIDEFace::m_fNormalsRenderLength = (FLOAT)0.2;

CListInt *CSLIDEFace::m_plistVertices = NULL;
UINT_32 CSLIDEFace::m_uiCount = 0;

BOOL CSLIDEFace::m_bScanConvertEnabled = TRUE;
CScanConverter *CSLIDEFace::m_pCScanConverter = NULL;


//////////////////////////////////////////////////////////////////////////////
// CSLIDEFace Class
//

CSLIDEFace::CSLIDEFace()
{
  m_uiCount++;
  if ( m_plistVertices == NULL )
    {
      m_plistVertices = new CListInt;
      ASSERT( m_plistVertices != NULL );
      m_plistVertices->Init();
    }

  m_pcSurfaceID = NULL;
  m_pCSLIDESurface = NULL;
  m_plistPointIDs = NULL;
}

CSLIDEFace::~CSLIDEFace()
{
  m_uiCount--;
  if ( m_uiCount == 0 )
    {
      if ( m_plistVertices != NULL )
	{
	  delete m_plistVertices;
	  m_plistVertices = NULL;
	}
    }
}

VOID CSLIDEFace::Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, CHAR *pcID, 
		      CListInt *plistPointIDs)
{
  CSLIDEPrimitive::Init(pcFile, uiLine, uiColumn, pcID);

  ASSERT( plistPointIDs != NULL );

  m_pcSurfaceID = gcppcSLIDEReservedWords[sliderwtSLF_INHERIT];
  m_pCSLIDESurface = NULL;

  m_plistPointIDs = plistPointIDs;
  m_listCSLIDEPoint.Init();

  SetAutoNormals(FALSE);
  SetAutoNormalsDynamic(FALSE);
}

VOID CSLIDEFace::Uninit()
{
  CSLIDEPrimitive::Uninit();

  if ( m_plistPointIDs != NULL )
    {
      delete m_plistPointIDs;
      m_plistPointIDs = NULL;
    }
}

BOOL CSLIDEFace::Initialized()
{
  if ( !CSLIDEPrimitive::Initialized() )
    {
      return FALSE;
    }

  return ( m_pcSurfaceID != NULL &&
	   m_plistPointIDs != NULL );
}

BOOL CSLIDEFace::CheckSemantics(CSLIDEWorld *pCSLIDEWorld)
{
  BOOL bCorrect;
  CListIterInt iter;
  CHAR *pcID;
  CSLIDEPoint *pCSLIDEPoint;

  if ( !CSLIDEPrimitive::CheckSemantics(pCSLIDEWorld) )
    {
      return FALSE;
    }

  bCorrect = TRUE;

  // Check Surface Reference

  m_pCSLIDESurface = pCSLIDEWorld->PeekSurface(m_pcSurfaceID);
  if ( m_pCSLIDESurface == NULL )
    {
      HandleError("Undefined %s \"%s\"", 
		  gcppcSLIDEReservedWords[sliderwtSurface],
		  m_pcSurfaceID);
      bCorrect = FALSE;
    }

  // Check Point References

  iter.Init(m_plistPointIDs);

  for ( pcID = (CHAR*)(iter.PeekFirst());
	pcID != NULL;
	pcID = (CHAR*)(iter.PeekNext()) )
    {
      pCSLIDEPoint = pCSLIDEWorld->PeekPoint(pcID);
      if ( pCSLIDEPoint == NULL )
	{
	  HandleError("Undefined %s \"%s\"", gcppcSLIDEReservedWords[sliderwtPoint], 
		      pcID);
	  bCorrect = FALSE;
	}
      m_listCSLIDEPoint.InsertLast(pCSLIDEPoint, 0);
    }

  return bCorrect;
}

BOOL CSLIDEFace::Valid()
{
  if ( !CSLIDEPrimitive::Valid() )
    {
      return FALSE;
    }

  return (GetPointCount() >= 3);
}

CSLIDEPrimitiveIter *CSLIDEFace::AllocateIter()
{
  CSLIDEFaceIter *pCSLIDEFaceIter;

  pCSLIDEFaceIter = new CSLIDEFaceIter;
  ASSERT(pCSLIDEFaceIter != NULL );
  pCSLIDEFaceIter->Init(this);

  return pCSLIDEFaceIter;
}

VOID CSLIDEFace::FreeIter(CSLIDEPrimitiveIter *pCSLIDEFaceIter)
{
  delete pCSLIDEFaceIter;
}

VOID CSLIDEFace::InsertIntoBound(CBound *pCBound)
{
  CListIterInt iter;
  CSLIDEPoint *pCSLIDEPoint;

  ASSERT( pCBound != NULL );

  iter.Init(&m_listCSLIDEPoint);

  for ( pCSLIDEPoint = (CSLIDEPoint*)iter.PeekFirst(); 
	pCSLIDEPoint != NULL;
	pCSLIDEPoint = (CSLIDEPoint*)iter.PeekNext() )
    {
      pCSLIDEPoint->InsertIntoBound(pCBound);
    }
}

VOID CSLIDEFace::Preprocess()
{
  CheckDynamic();
  UpdateState();

  // All faces must calculate their normal at least once

  CalculateNormal();
  CalculateCentroid();
  
  // Calculate the D coefficient of the Tangent Plane defined by the normal
  m_vNormal.CompletePlaneEquation(m_ptCentroid);
}

//////////////////////////////////////////////////////////////////////////////
// START: ASSIGNMENT 8: Deal with dynamic Point normals 
//

// This should only be called by objects which can be smooth shaded
VOID CSLIDEFace::PreprocessAutoNormals()
{
  if ( !GetAutoNormals() )
    {
      // We do not contribute to any vertex AutoNormals because we are not
      // part of any smooth shaded objects
      return;
    }

  // Calculate Point Normals
}

// This should only be called by objects which can be smooth shaded
// Check to see if any of the face's points has dynamic auto normals
// in which case we must remember to resend our normal to it on every update
VOID CSLIDEFace::CheckAutoNormalsDynamic()
{
  // I did not remove anything from here!

  CListIterInt iter;
  CSLIDEPoint *pCSLIDEPoint;

  iter.Init(&m_listCSLIDEPoint);

  for ( pCSLIDEPoint = (CSLIDEPoint*)iter.PeekFirst(); 
	pCSLIDEPoint != NULL;
	pCSLIDEPoint = (CSLIDEPoint*)iter.PeekNext() )
    {
      if ( pCSLIDEPoint->GetAutoNormalDynamic() )
	{
	  SetAutoNormalsDynamic(TRUE);
	  SetDynamic(TRUE);
	  break;
	}
    }
}
//
// END  : ASSIGNMENT 8: Deal with dynamic Point normals 
//////////////////////////////////////////////////////////////////////////////

VOID CSLIDEFace::CheckDynamic()
{
  CListIterInt iter;
  CSLIDEPoint *pCSLIDEPoint;
  BOOL bBounded;
  BOOL bBoundDynamic;

  bBounded = GetDynamic();
  bBoundDynamic = GetDynamic();

  iter.Init(&m_listCSLIDEPoint);

  for ( pCSLIDEPoint = (CSLIDEPoint*)iter.PeekFirst(); 
	pCSLIDEPoint != NULL;
	pCSLIDEPoint = (CSLIDEPoint*)iter.PeekNext() )
    {
      bBounded = TRUE;
      if ( pCSLIDEPoint->GetBoundDynamic() )
	{
	  bBoundDynamic = TRUE;
	  break;
	}
    }

  SetBounded(bBounded);
  SetBoundDynamic(bBoundDynamic);
  SetDynamic(bBoundDynamic);
  SetAutoNormalsDynamic(FALSE);
  ASSERT( !(!GetBounded() && GetBoundDynamic()) );
}

//////////////////////////////////////////////////////////////////////////////
// START: ASSIGNMENT 8: Deal with dynamic Point normals 
//
VOID CSLIDEFace::UpdateState()
{
  if ( GetBoundDynamic() )
    {
      CalculateNormal();
      CalculateCentroid();

      // Calculate the D coefficient of the Tangent Plane defined by the normal
      m_vNormal.CompletePlaneEquation(m_ptCentroid);
    }

  if ( GetAutoNormalsDynamic() )
    {
      // Calculate Point Normals

      // ???

    }
}
//
// END  : ASSIGNMENT 8: Deal with dynamic Point normals 
//////////////////////////////////////////////////////////////////////////////

// This uses Newell's Algorithm for finding the best fit planar equation
// for a set of points in 3 space (pg. 476 Foley and van Dam)
// I modified it so that the normal would be positive for counter-clockwise
// polygons, and I removed the 1/2 coefficient because we are concerned 
// with relative areas, not absolute areas.

VOID CSLIDEFace::CalculateNormal()
{
  CListIterInt iter;
  CSLIDEPoint *pCSLIDEPoint;
  Point ptStart, ptPrev, ptCurr;

  m_vNormal.Null();

  iter.Init(&m_listCSLIDEPoint);

  pCSLIDEPoint = (CSLIDEPoint*)iter.PeekFirst(); 
  if ( pCSLIDEPoint == NULL )
    {
      HandleWarning("Null normal because no points for face");
      return;
    }
  ptStart = pCSLIDEPoint->GetPoint();
  ptPrev = ptStart;

  for ( pCSLIDEPoint = (CSLIDEPoint*)iter.PeekNext(); 
	pCSLIDEPoint != NULL;
	pCSLIDEPoint = (CSLIDEPoint*)iter.PeekNext() )
    {
      ptCurr = pCSLIDEPoint->GetPoint();

      m_vNormal[X] += (ptPrev[Z] + ptCurr[Z])*(ptPrev[Y] - ptCurr[Y]);
      m_vNormal[Y] += (ptPrev[X] + ptCurr[X])*(ptPrev[Z] - ptCurr[Z]);
      m_vNormal[Z] += (ptPrev[Y] + ptCurr[Y])*(ptPrev[X] - ptCurr[X]);

      ptPrev = ptCurr;
    }

  ptCurr = ptStart;
  m_vNormal[X] += (ptPrev[Z] + ptCurr[Z])*(ptPrev[Y] - ptCurr[Y]);
  m_vNormal[Y] += (ptPrev[X] + ptCurr[X])*(ptPrev[Z] - ptCurr[Z]);
  m_vNormal[Z] += (ptPrev[Y] + ptCurr[Y])*(ptPrev[X] - ptCurr[X]);
  
  if ( !m_vNormal.Normalize() )
    {
      HandleWarning("Null normal because of zero area face");
    }
}

VOID CSLIDEFace::CalculateCentroid()
{
  CListIterInt iter;
  UINT_32 i;
  CSLIDEPoint *pCSLIDEPoint;

  iter.Init(&m_listCSLIDEPoint);

  m_ptCentroid.Null();

  i = 0;
  for ( pCSLIDEPoint = (CSLIDEPoint*)iter.PeekFirst();
	pCSLIDEPoint != NULL;
	pCSLIDEPoint = (CSLIDEPoint*)iter.PeekNext() )
    {
      m_ptCentroid = m_ptCentroid + pCSLIDEPoint->GetPoint();
      i++;
    }

  if ( i > 0 )
    {
      m_ptCentroid = m_ptCentroid / (FLOAT)i;
    }

  m_ptCentroid[W] = 1.0;;
}

//////////////////////////////////////////////////////////////////////////////
// START: ASSIGNMENT 8: Deal with Lighting and Shading
//
VOID CSLIDEFace::Render(const Matrix &mGvp_proj, 
			const Matrix &mGproj_vrc, 
			const Matrix &mGvrc_proj, 
			const Matrix &mGvrc_obj, 
			const Matrix &mGobj_vrc, 
			CBound *pCBoundClip, 
			CListInt *plistCSLIDELightPath,
			CSLIDESurface *pCSLIDESurface,
			SLIDEShadingType eShading, SLIDELODType eLOD, 
			SLIDESolidType eSolid)
{
  GLenum mode;
  CListIterInt iter;
  CSLIDEPoint *pCSLIDEPoint;
  Matrix mGproj_obj;
  Vertex *pvx;

  // Check Back Face Culling
  if ( GetBackFaceCullingEnabled() && eSolid == slidestSLF_SOLID )
    {
      Vector vNormal;
      
      //This is the correct line
      // Doing row multiplication instead of transposing the matrix
      vNormal = m_vNormal * mGobj_vrc * mGvrc_proj; 

      // This is the wrong way to transform normals to prove that 
      // all this Matrix Inverse stuff was really necessary
      //vNormal = m_vNormal;
      //vNormal[W] = 0.0;
      //vNormal = mGproj_vrc * mGvrc_obj * vNormal; 
      //vNormal = mGvrc_obj * vNormal; 

      if ( vNormal[Z] <= 0.0 )
	{
	  return;
	}
    }

  // Modify Rendering Properties
  ASSERT( eLOD == slidelodtSLF_EDGES || eLOD == slidelodtSLF_FULL );
  ModifySurfacePointer(pCSLIDESurface);

  // Modify Shading based on LOD ???

  // Build a list of Vertices which we dynamically allocate
  // this will be the input to the clipping routine
  iter.Init(&m_listCSLIDEPoint);

  for ( pCSLIDEPoint = (CSLIDEPoint*)iter.PeekFirst(); 
	pCSLIDEPoint != NULL;
	pCSLIDEPoint = (CSLIDEPoint*)iter.PeekNext() )
    {
      pvx = CBound::AllocateVertex();
      ASSERT( pvx != NULL );

      pvx->SetPoint() = pCSLIDEPoint->GetPoint();
      pvx->SetNormal() = pCSLIDEPoint->GetNormal();
      pvx->SetSurface(pCSLIDESurface);
      pvx->SetColor() = pvx->GetSurface()->GetColor();
      pvx->SetTexture() = pCSLIDEPoint->GetTexture();

      m_plistVertices->InsertLast(pvx, 0);
    }

  // Color the Vertices of the polygon
  switch ( eShading )
    {
    case slideshtSLF_WIRE:
      { 
      }
      break;

    case slideshtSLF_INHERIT:
    case slideshtSLF_FLAT:
      { 
	Vector vColor;
	
	if ( GetRenderNormalsEnabled() )
	  {
	    gilPushAttrib(GL_CURRENT_BIT);
	    {
	      gilColor3f(1.0, 0.0, 1.0);
	      gilBegin(GL_LINES);
	      {
		RenderNormal(mGvp_proj, mGproj_vrc, mGvrc_obj, mGobj_vrc, 
			     m_ptCentroid, m_vNormal);
	      }
	      gilEnd();
	    }
	    gilPopAttrib();
	  }

	// All the vertices get the same color as the polygon
	if ( plistCSLIDELightPath == NULL )
	  {
	    vColor = pCSLIDESurface->GetColor();
	  }
	else
	  {
	    CSLIDELightPath *pCSLIDELightPath;
	    Point ptCentroid;
	    Vector vNormal;
	    
	    // Transform the Centroid and the Normal to the VRC system
	    ptCentroid = mGvrc_obj * m_ptCentroid;
	    vNormal = m_vNormal * mGobj_vrc;

	    vColor.Null();

	    iter.Init(plistCSLIDELightPath);

	    for ( pCSLIDELightPath = (CSLIDELightPath*)iter.PeekFirst();
	          pCSLIDELightPath != NULL;
		  pCSLIDELightPath = (CSLIDELightPath*)iter.PeekNext() )
	      {
		vColor += pCSLIDELightPath->GetLight()->
		  CalculateColor(ptCentroid,
				 vNormal, 
				 pCSLIDESurface,
				 pCSLIDELightPath->GetGlight_vrc(),
				 pCSLIDELightPath->GetGvrc_light());
	      }
	  }

	iter.Init(m_plistVertices);
	
	for ( pvx = (Vertex*)iter.PeekFirst(); 
	      pvx != NULL;
	      pvx = (Vertex*)iter.PeekNext() )
	  {
	    pvx->SetColor() = vColor;
	  }
      }
      break;

    case slideshtSLF_GOURAUD:
    case slideshtSLF_PHONG: // I am not supporting Phong Shading yet
      {
	if ( GetRenderNormalsEnabled() )
	  {
	    gilPushAttrib(GL_CURRENT_BIT);
	    {
	      gilColor3f(1.0, 0.0, 1.0);
	      gilBegin(GL_LINES);
	      {
		iter.Init(m_plistVertices);
		
		for ( pvx = (Vertex*)iter.PeekFirst(); 
		      pvx != NULL;
		      pvx = (Vertex*)iter.PeekNext() )
		  {
		    RenderNormal(mGvp_proj, mGproj_vrc, mGvrc_obj, mGobj_vrc, 
				 pvx->GetPoint(), pvx->GetNormal());
		  }
	      }
	      gilEnd();
	    }
	    gilPopAttrib();
	  }

	if ( plistCSLIDELightPath != NULL )
	  {
	    // Calculate a color value for each vertex
	  }
      }
      break;

    default:
      ASSERT( FALSE );
      break;
    }
  
  // From this point on the Vertex::Normal does not matter

  iter.Init(m_plistVertices);
  
  mGproj_obj = mGproj_vrc * mGvrc_obj;

  // Transform the points to the Warped Projective space
  for ( pvx = (Vertex*)iter.PeekFirst(); 
	pvx != NULL;
	pvx = (Vertex*)iter.PeekNext() )
    {
      pvx->SetPoint() = mGproj_obj * pvx->GetPoint();
    }
  

  // Clip the polygon against the 4D bound if clipping is enabled
  // and we do not lie entirely within the camera's view frustum
  if ( GetClipEnabled() && pCBoundClip != NULL )
    {
      // The clipping routine should modify the list of vertices.
      // It should deallocate any vertices it removes by calling 
      // CBound::DeallocateVertex(pvx) on them.  It should also
      // insert any new vertices into the m_plistVertices, so that
      // when the ClipPolygon routine finishes the list is the clipped
      // polygon contour.
      // This should correctly interpolate the color values.
      pCBoundClip->ClipPolygon(*m_plistVertices);
    }

  // Transform the clipped vertices into pixel coords, homogenize them, and 
  // render them. We remove the vertices from the list of vertices and 
  // deallocate these temporary vertex structures as we go.
  // Homogenize the Clipped Points and transform to pixel coords
  // and render

  if ( GetScanConvertEnabled() )
    {
      // Send in vertices in 4D before the division by W, but
      // scaled to the viewport.  Vertices contain the correct vertex coloring.
      GetScanConverter()->FillPolygon(mGvp_proj, *m_plistVertices);
      for ( pvx = (Vertex*)m_plistVertices->RemoveFirst(); 
	    pvx != NULL;
	    pvx = (Vertex*)m_plistVertices->RemoveFirst() )
	{
	  CBound::DeallocateVertex(pvx);
	}
    }
  else
    {
      if ( eShading == slideshtSLF_WIRE )
	{
	  mode = GL_LINE_LOOP;
	}
      else
	{
	  mode = GL_POLYGON;
	}
      gilBegin(mode);
      {
	for ( pvx = (Vertex*)m_plistVertices->RemoveFirst(); 
	      pvx != NULL;
	      pvx = (Vertex*)m_plistVertices->RemoveFirst() )
	  {
	    pvx->SetPoint().Homogenize();
	    pvx->SetPoint() = mGvp_proj * pvx->GetPoint();
	    
	    pvx->Render();
	    CBound::DeallocateVertex(pvx);
	  }
      }
      gilEnd();
    }
}
//
// END  : ASSIGNMENT 8: Deal with Lighting and Shading
//////////////////////////////////////////////////////////////////////////////

VOID CSLIDEFace::ModifySurfacePointer(CSLIDESurface* &pCSLIDESurface)
{
  if ( m_pCSLIDESurface != CSLIDESurface::GetSLF_INHERIT() )
    {
      pCSLIDESurface = m_pCSLIDESurface;
    }
}

VOID CSLIDEFace::RenderNormal(const Matrix &mGvp_proj, const Matrix &mGproj_vrc, 
			      const Matrix &mGvrc_obj, const Matrix &mGobj_vrc,
			      const Point &pt, const Vector &vNormal)
{
  Point ptStart, ptEnd;
  Vector v;
  Matrix m;

  v = vNormal * mGobj_vrc;
  ptStart = mGvrc_obj * pt;

  v[W] = 0;
  v.Normalize();
  ptEnd = ptStart + GetNormalsRenderLength() * v;

  ptStart = mGproj_vrc * ptStart;
  ptEnd = mGproj_vrc * ptEnd;

  ptStart.Homogenize();
  ptEnd.Homogenize();

  ptStart = mGvp_proj * ptStart;
  ptEnd = mGvp_proj * ptEnd;

  gilVertex3fv(ptStart.Get());
  gilVertex3fv(ptEnd.Get());
}

VOID CSLIDEFace::WriteSLIDE(FILE *fpOutputFile)
{
  CListIterInt iter;
  CSLIDEPoint *pCSLIDEPoint;

  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "%s %s ( ", gcppcSLIDEReservedWords[sliderwtFace], 
	  m_pcID);

  iter.Init(&m_listCSLIDEPoint);

  for ( pCSLIDEPoint = (CSLIDEPoint*)iter.PeekFirst(); 
	pCSLIDEPoint != NULL;
	pCSLIDEPoint = (CSLIDEPoint*)iter.PeekNext() )
    {
      fprintf(fpOutputFile, "%s ", pCSLIDEPoint->GetID());      
    }

  fprintf(fpOutputFile, ")\n");

  if ( m_pcSurfaceID != gcppcSLIDEReservedWords[sliderwtSLF_INHERIT] )
    {
      fprintf(fpOutputFile, "  %s %s\n", 
	      gcppcSLIDEReservedWords[sliderwtSurface], 
	      m_pcSurfaceID);
    } 

  fprintf(fpOutputFile, "%s\n", 
	  gcppcSLIDEReservedWords[sliderwtEndFace]);
}

UINT_32 CSLIDEFace::GetPointCount()
{
  return m_listCSLIDEPoint.GetLength();
}

//////////////////////////////////////////////////////////////////////////////
// CSLIDEFaceIter Class
//

CSLIDEFaceIter::CSLIDEFaceIter()
{
}

CSLIDEFaceIter::~CSLIDEFaceIter()
{
}

VOID CSLIDEFaceIter::Init(CSLIDEFace *pCSLIDEFace)
{
  CSLIDEPrimitiveIter::Init(pCSLIDEFace);

  m_iter.Init(&(pCSLIDEFace->m_listCSLIDEPoint));
}

VOID CSLIDEFaceIter::Uninit()
{
  CSLIDEPrimitiveIter::Uninit();
}

CSLIDEPoint *CSLIDEFaceIter::PeekFirstCSLIDEPoint()
{
  return (CSLIDEPoint*)(m_iter.PeekFirst());
}

CSLIDEPoint *CSLIDEFaceIter::PeekNextCSLIDEPoint()
{
  return (CSLIDEPoint*)(m_iter.PeekNext());
}

CSLIDEPoint *CSLIDEFaceIter::PeekLastCSLIDEPoint()
{
  return (CSLIDEPoint*)(m_iter.PeekLast());
}

CSLIDEPoint *CSLIDEFaceIter::PeekPrevCSLIDEPoint()
{
  return (CSLIDEPoint*)(m_iter.PeekPrev());
}
