#include "slideobject.h"

#include "slideworld.h"
#include "slidesurface.h"
#include "slideface.h"

//////////////////////////////////////////////////////////////////////
// CSLIDEObject Class
//

CSLIDEObject::CSLIDEObject()
{
  m_pcSurfaceID = NULL;
  m_pCSLIDESurface = NULL;
  m_dyniShading.SetValue(slideshtSLF_WIRE);
  m_dyniLOD.SetValue(slidelodtSLF_FULL);
  m_dyniSolid.SetValue(slidestSLF_SOLID);
  m_plistFaceIDs = NULL;
}

CSLIDEObject::~CSLIDEObject()
{
}

VOID CSLIDEObject::Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, 
			CHAR *pcID,
			CListInt *plistFaceIDs)
{
  CSLIDENode::Init(pcFile, uiLine, uiColumn, pcID);

  m_pcSurfaceID = gcppcSLIDEReservedWords[sliderwtSLF_INHERIT];
  m_pCSLIDESurface = NULL;
  m_dyniShading.SetValue(slideshtSLF_INHERIT);
  m_dyniLOD.SetValue(slidelodtSLF_FULL);
  m_dyniSolid.SetValue(slidestSLF_SOLID);
  m_plistFaceIDs = plistFaceIDs;
  m_listCSLIDEFace.Init();
}

VOID CSLIDEObject::Uninit()
{
  CSLIDENode::Uninit();

  m_pcSurfaceID = NULL;
  m_pCSLIDESurface = NULL;

  if ( m_plistFaceIDs != NULL )
    {
      delete m_plistFaceIDs;
      m_plistFaceIDs = NULL;
    }
}

BOOL CSLIDEObject::Initialized()
{
  if ( !CSLIDENode::Initialized() )
    {
      return FALSE;
    }

  return (m_plistFaceIDs != NULL);
}

BOOL CSLIDEObject::CheckSemantics(CSLIDEWorld *pCSLIDEWorld)
{
  BOOL bCorrect;
  CListIterInt iter;
  CHAR *pcID;
  CSLIDEFace *pCSLIDEFace;

  if ( !CSLIDENode::CheckSemantics(pCSLIDEWorld) )
    {
      return FALSE;
    }

  bCorrect = TRUE;

  // Check Surface Reference

  m_pCSLIDESurface = pCSLIDEWorld->PeekSurface(m_pcSurfaceID);
  if ( m_pCSLIDESurface == NULL )
    {
      HandleError("Undefined %s \"%s\"", 
		  gcppcSLIDEReservedWords[sliderwtSurface],
		  m_pcSurfaceID);
      bCorrect = FALSE;
    }

  // Check Shading Type

  ASSERT( 0 <= m_dyniShading.GetValue() &&
	  m_dyniShading.GetValue() < slideshtSize );

  // Check LOD Type

  ASSERT( 0 <= m_dyniLOD.GetValue() &&
	  m_dyniLOD.GetValue() < slidelodtSize );

  // Check Solid Type

  ASSERT( 0 <= m_dyniSolid.GetValue() &&
	  m_dyniSolid.GetValue() < slidestSize );

  // Check Face References

  iter.Init(m_plistFaceIDs);

  for ( pcID = (CHAR*)(iter.PeekFirst());
	pcID != NULL;
	pcID = (CHAR*)(iter.PeekNext()) )
    {
      pCSLIDEFace = pCSLIDEWorld->PeekFace(pcID);
      if ( pCSLIDEFace == NULL )
	{
	  HandleError("Undefined %s \"%s\"", gcppcSLIDEReservedWords[sliderwtFace], pcID);
	  bCorrect = FALSE;
	}
      m_listCSLIDEFace.InsertLast(pCSLIDEFace, 0);
    }

  return bCorrect;
}

BOOL CSLIDEObject::Valid()
{
  if ( !CSLIDENode::Valid() )
    {
      return FALSE;
    }

  return ( ( 0 <= m_dyniSolid.GetValue() &&
	     m_dyniSolid.GetValue() < slidestSize ) &&
	   ( 0 <= m_dyniShading.GetValue() &&
	     m_dyniShading.GetValue() < slideshtSize ) &&
	   ( 0 <= m_dyniLOD.GetValue() &&
	     m_dyniLOD.GetValue() < slidelodtSize ) );
	   //GetFaceCount() > 0 );
}

CSLIDEPrimitiveIter *CSLIDEObject::AllocateIter()
{
  CSLIDEObjectIter *pCSLIDEObjectIter;

  pCSLIDEObjectIter = new CSLIDEObjectIter;
  ASSERT(pCSLIDEObjectIter != NULL );
  pCSLIDEObjectIter->Init(this);

  return pCSLIDEObjectIter;
}

VOID CSLIDEObject::FreeIter(CSLIDEPrimitiveIter *pCSLIDEObjectIter)
{
  delete pCSLIDEObjectIter;
}

CSLIDECamera *CSLIDEObject::FindCamera(CListIterInt *piterIDs, 
				       CListInt *plistCSLIDEInstance)
{
  ASSERT( piterIDs != NULL );
  ASSERT( plistCSLIDEInstance != NULL );

  HandleError("Bad %s path: %s \"%s\" encountered", 
	      gcppcSLIDEReservedWords[sliderwtCamera],
	      gcppcSLIDEReservedWords[sliderwtObject], GetID());

  return NULL;
}

CSLIDELight *CSLIDEObject::FindLight(CListIterInt *piterIDs, 
				     CListInt *plistCSLIDEInstance)
{
  ASSERT( piterIDs != NULL );
  ASSERT( plistCSLIDEInstance != NULL );

  HandleError("Bad %s path: %s \"%s\" encountered", 
	      gcppcSLIDEReservedWords[sliderwtLight],
	      gcppcSLIDEReservedWords[sliderwtObject], GetID());

  return NULL;
}

VOID CSLIDEObject::Preprocess()
{
  UpdateState();
  CheckDynamic();
  // This sets BoundEmpty(TRUE)
  m_bound.Null();
  SetBoundChanged(FALSE);
  SetBoundValid(FALSE);
}

VOID CSLIDEObject::CheckDynamic()
{
  SetDynamic(m_dyniLOD.GetDynamic() ||
	     ((m_dyniLOD.GetValue() != slidelodtSLF_OFF) &&
	      (m_dyniShading.GetDynamic() || m_dyniSolid.GetDynamic())));
}

//////////////////////////////////////////////////////////////////////////////
// START: ASSIGNMENT 8: Handle dynamic point normals
//
VOID CSLIDEObject::PreprocessGraph()
{
  if ( !GetBoundValid() )
    {
      CListIterInt iter;
      CSLIDEFace *pCSLIDEFace;
      BOOL bBounded;
      BOOL bBoundDynamic;
      BOOL bAutoNormals;
      
      bBounded = FALSE;
      bBoundDynamic = FALSE;

      iter.Init(&m_listCSLIDEFace);
      
      for ( pCSLIDEFace = (CSLIDEFace*)iter.PeekFirst(); 
	    pCSLIDEFace != NULL;
	    pCSLIDEFace = (CSLIDEFace*)iter.PeekNext() )
	{
	  bBounded = bBounded || pCSLIDEFace->GetBounded();
	  bBoundDynamic = bBoundDynamic || pCSLIDEFace->GetBoundDynamic();
	}
      
      SetBounded(bBounded && 
		 (m_dyniLOD.GetDynamic() || 
		  (m_dyniLOD.GetValue() != slidelodtSLF_OFF)));
      SetBoundDynamic(bBounded && 
		      (m_dyniLOD.GetDynamic() || 
		       ((m_dyniLOD.GetValue() != slidelodtSLF_OFF) && 
			bBoundDynamic)));
      ASSERT( !(!GetBounded() && GetBoundDynamic()) );
      
      if ( m_dyniLOD.GetValue() == slidelodtSLF_OFF )
	{
	  m_bound.Null();
	}
      else
	{
	  CalculateBound();
	}
      SetBoundChanged(FALSE);
      SetBoundValid(TRUE);
    }
}
//
// END  : ASSIGNMENT 8: Handle dynamic point normals
//////////////////////////////////////////////////////////////////////////////

VOID CSLIDEObject::Update()
{
  SetBoundChanged(FALSE);
  if ( GetDynamic() )
    {
      UpdateState();
    }

  if ( GetBoundDynamic() )
    {
      SetBoundValid(FALSE);
    }
}

VOID CSLIDEObject::UpdateState()
{
  m_dyniSolid.Update();
  if ( !( 0 <= m_dyniSolid.GetValue() &&
	  m_dyniSolid.GetValue() < slidestSize ) )
    {
      HandleError("Illegal solid flag value:  %ld", 
		  m_dyniSolid.GetValue());
      m_dyniSolid.SetValue(slidestSLF_HOLLOW);
    }

  m_dyniShading.Update();
  if ( !( 0 <= m_dyniShading.GetValue() &&
	  m_dyniShading.GetValue() < slideshtSize ) )
    {
      HandleError("Illegal Shading flag value:  %ld", 
		  m_dyniShading.GetValue());
      m_dyniShading.SetValue(slideshtSLF_FLAT);
    }

  m_dyniLOD.Update();
  if ( !( 0 <= m_dyniLOD.GetValue() &&
	  m_dyniLOD.GetValue() < slideshtSize ) )
    {
      HandleError("Illegal LOD flag value:  %ld", 
		  m_dyniLOD.GetValue());
      m_dyniLOD.SetValue(slidelodtSLF_FULL);
    }
}

VOID CSLIDEObject::UpdateGraph()
{
  if ( !GetBoundValid() )
    {
      if ( m_dyniLOD.GetValue() == slidelodtSLF_OFF )
	{
	  if ( !m_bound.GetEmpty() )
	    {
	      SetBoundChanged(TRUE);
	      m_bound.Null();
	    }
	}
      else
	{
	  CalculateBound();
	}
      SetBoundValid(TRUE);
    }
}

VOID CSLIDEObject::InsertIntoBound(const Matrix &m, CBound &bound)
{
  bound.Insert(m, m_bound);
}

VOID CSLIDEObject::CalculateBound()
{
  CListIterInt iter;
  CSLIDEFace *pCSLIDEFace;

  iter.Init(&m_listCSLIDEFace);
  
  if ( GetCheckBoundChanged() )
    {
      CBound bound;
      
      for ( pCSLIDEFace = (CSLIDEFace*)iter.PeekFirst(); 
	    pCSLIDEFace != NULL;
	    pCSLIDEFace = (CSLIDEFace*)iter.PeekNext() )
	{
	  pCSLIDEFace->InsertIntoBound(&bound);
	}
      
      if ( bound != m_bound )
	{
	  m_bound = bound;
	  SetBoundChanged(TRUE);
	}
    }
  else
    {
      m_bound.Null();
      for ( pCSLIDEFace = (CSLIDEFace*)iter.PeekFirst(); 
	    pCSLIDEFace != NULL;
	    pCSLIDEFace = (CSLIDEFace*)iter.PeekNext() )
	{
	  pCSLIDEFace->InsertIntoBound(&m_bound);
	}
      
      SetBoundChanged(TRUE);
    }
}

VOID CSLIDEObject::ModifySurfacePointer(CSLIDESurface* &pCSLIDESurface)
{
  if ( m_pCSLIDESurface != CSLIDESurface::GetSLF_INHERIT() )
    {
      pCSLIDESurface = m_pCSLIDESurface;
    }
}

//////////////////////////////////////////////////////////////////////////////
// START: ASSIGNMENT 8: Handle Shading Flags
//
VOID CSLIDEObject::ModifyShadingFlag(SLIDEShadingType &eShading)
{
}
//
// END  : ASSIGNMENT 8: Handle Shading Flags
//////////////////////////////////////////////////////////////////////////////

VOID CSLIDEObject::ModifyLODFlag(SLIDELODType &eLOD)
{
  eLOD = (SLIDELODType)MIN((INT_32)eLOD, m_dyniLOD.GetValue());
}

CBound::BoundIntersectionType CSLIDEObject::CheckBoundReject(const Matrix &m,
							     CBound *pCBoundClip)
{
  ASSERT( pCBoundClip != NULL );

  return pCBoundClip->Intersect(m, m_bound);
}

VOID CSLIDEObject::RenderGeometry(const Matrix &mGvp_proj, 
				  const Matrix &mGproj_vrc, 
				  const Matrix &mGvrc_proj, 
				  const Matrix &mGvrc_obj, 
				  const Matrix &mGobj_vrc, 
				  CBound *pCBoundClip, 
				  CListInt *plistLights,
				  CSLIDESurface *pCSLIDESurface,
				  SLIDEShadingType eShading, 
				  SLIDELODType eLOD)
{
  CListIterInt iter;
  CSLIDEFace *pCSLIDEFace;
  
  //m_pCSLIDESurface->Push();
  {
    iter.Init(&m_listCSLIDEFace);
    
    for ( pCSLIDEFace = (CSLIDEFace*)iter.PeekFirst(); 
	  pCSLIDEFace != NULL;
	  pCSLIDEFace = (CSLIDEFace*)iter.PeekNext() )
      {
	pCSLIDEFace->Render(mGvp_proj, mGproj_vrc, mGvrc_proj, 
			    mGvrc_obj, mGobj_vrc, 
			    pCBoundClip, plistLights, pCSLIDESurface,
			    eShading, eLOD, 
			    (SLIDESolidType)m_dyniSolid.GetValue());
      }
  }
  //m_pCSLIDESurface->Pop();
}

VOID CSLIDEObject::RenderBound(const Matrix &mGvp_proj, const Matrix &mGproj_vrc, 
			       const Matrix &mGvrc_obj)
{
  gilPushAttrib(GL_CURRENT_BIT);
  {
    if ( GetBoundChanged() && GetColorChangedBounds() )
      {
	// Orange for Changed Bounding Boxes
	gilColor3f(1.0, 0.373, 0.0);
      }
    else if ( GetBoundDynamic() && GetColorDynamicBounds() )
      {
	// Red for Dynamic Bounding Boxes
	gilColor3f(1.0, 0.0, 0.0);
      }
    else
      {
        // White for Static Bounding Boxes
	gilColor3f(1.0, 1.0, 1.0);
      }
    m_bound.Render(mGvp_proj, mGproj_vrc * mGvrc_obj);
  }
  gilPopAttrib();
}

VOID CSLIDEObject::WriteSLIDE(FILE *fpOutputFile)
{
  CListIterInt iter;
  CSLIDEFace *pCSLIDEFace;

  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "%s %s ( ", 
	  gcppcSLIDEReservedWords[sliderwtObject], m_pcID);

  iter.Init(&m_listCSLIDEFace);

  for ( pCSLIDEFace = (CSLIDEFace*)iter.PeekFirst(); 
	pCSLIDEFace != NULL;
	pCSLIDEFace = (CSLIDEFace*)iter.PeekNext() )
    {
      fprintf(fpOutputFile, "%s ", pCSLIDEFace->GetID());      
    }

  fprintf(fpOutputFile, ")\n");

  if ( m_pcSurfaceID != gcppcSLIDEReservedWords[sliderwtSLF_INHERIT] )
    {
      fprintf(fpOutputFile, "  %s %s\n",
	      gcppcSLIDEReservedWords[sliderwtSurface], 
	      m_pcSurfaceID);
    } 
  if ( m_dyniShading.GetValue() != slideshtSLF_INHERIT )
    {
      fprintf(fpOutputFile, "  %s %s\n",
	      gcppcSLIDEReservedWords[sliderwtShading], 
	      gcppcSLIDEShadingReservedWords[m_dyniShading.GetValue()]);
    }
  if ( m_dyniLOD.GetValue() != slidelodtSLF_FULL )
    {
      fprintf(fpOutputFile, "    %s %s\n",
	      gcppcSLIDEReservedWords[sliderwtLOD], 
	      gcppcSLIDELODReservedWords[m_dyniLOD.GetValue()]);
    }
  if ( m_dyniSolid.GetValue() != slidestSLF_SOLID )
    {
      fprintf(fpOutputFile, "  %s %s\n",
	      gcppcSLIDEReservedWords[sliderwtSolid], 
	      gcppcSLIDESolidReservedWords[m_dyniSolid.GetValue()]);
    }

  fprintf(fpOutputFile, "%s\n", 
	  gcppcSLIDEReservedWords[sliderwtEndObject]);
}

UINT_32 CSLIDEObject::GetFaceCount()
{
  return m_listCSLIDEFace.GetLength();
}

//////////////////////////////////////////////////////////////////////////////
// CSLIDEObjectIter Class
//

CSLIDEObjectIter::CSLIDEObjectIter()
{
}

CSLIDEObjectIter::~CSLIDEObjectIter()
{
}

VOID CSLIDEObjectIter::Init(CSLIDEObject *pCSLIDEObject)
{
  CSLIDEPrimitiveIter::Init(pCSLIDEObject);

  m_iter.Init(&(pCSLIDEObject->m_listCSLIDEFace));
}

VOID CSLIDEObjectIter::Uninit()
{
  CSLIDEPrimitiveIter::Uninit();
}

CSLIDEFace *CSLIDEObjectIter::PeekFirstCSLIDEFace()
{
  return (CSLIDEFace*)(m_iter.PeekFirst());
}

CSLIDEFace *CSLIDEObjectIter::PeekNextCSLIDEFace()
{
  return (CSLIDEFace*)(m_iter.PeekNext());
}
