#ifndef __SLIDEPOINT_H__
#define __SLIDEPOINT_H__

#include "slideprimitive.h"

class CSLIDESurface;

//////////////////////////////////////////////////////////////////////
// CSLIDEPoint Class
//

class CSLIDEPoint : public CSLIDEPrimitive
{
//////////////////////////////////////////////////
// Enumerated Types
//

//////////////////////////////////////////////////
// Initialization
//
public:
  CSLIDEPoint();
  virtual ~CSLIDEPoint();

  virtual VOID Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, 
		    CHAR *pcID, 
		    DynPoint *pdynpt);
  virtual VOID Uninit();

  virtual BOOL Initialized();

  virtual BOOL CheckSemantics(CSLIDEWorld *pCSLIDEWorld);

  virtual BOOL Valid();

//////////////////////////////////////////////////
// Geometric Computations
//
public:
  VOID InsertIntoBound(CBound *pCBound);
//////////////////////////////////////////////////////////////////////////////
// START: ASSIGNMENT 8: Handle dynamic point normals
//
  VOID AddFaceNormal(const Vector &vNormal, FLOAT fWeight);

  VOID PreprocessNormal();
//
// END  : ASSIGNMENT 8: Handle dynamic point normals
//////////////////////////////////////////////////////////////////////////////

protected:
  VOID CheckDynamic();
//////////////////////////////////////////////////////////////////////////////
// START: ASSIGNMENT 8: Handle dynamic point normals
//
  VOID UpdateState();
//
// END  : ASSIGNMENT 8: Handle dynamic point normals
//////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////
// Rendering
//
public:
  virtual VOID Render(const Matrix &m);

//////////////////////////////////////////////////
// Output Functions
//
public:
  virtual VOID WriteSLIDE(FILE *fpOutputFile);

//////////////////////////////////////////////////
// Accessor Functions
//
public:
  inline VOID SetSurfaceID(CHAR *pcSurfaceID);
//////////////////////////////////////////////////////////////////////////////
// START: ASSIGNMENT 8: Handle dynamic point normals
//
  inline VOID SetNormal(DynVector *pdynvNormal);
//
// END  : ASSIGNMENT 8: Handle dynamic point normals
//////////////////////////////////////////////////////////////////////////////
  inline VOID SetTexture(DynPoint *pdynptTexture);

  inline CSLIDESurface *GetSurface();
  inline const Point &GetPoint();
  inline const Vector &GetNormal();
  inline const Point &GetTexture();

  inline BOOL GetBoundDynamic();
  inline VOID SetBoundDynamic(BOOL bBoundDynamic);
//////////////////////////////////////////////////////////////////////////////
// START: ASSIGNMENT 8: Handle dynamic point normals
//
  inline BOOL GetAutoNormal();
  inline VOID SetAutoNormal(BOOL bAutoNormal);
  inline BOOL GetAutoNormalDynamic();
  inline VOID SetAutoNormalDynamic(BOOL bAutoNormalDynamic);
//
// END  : ASSIGNMENT 8: Handle dynamic point normals
//////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////
// Class Variables
//
public:

//////////////////////////////////////////////////
// Member Variables
//
protected:
  CHAR *m_pcSurfaceID;
  CSLIDESurface *m_pCSLIDESurface;

  DynPoint m_pt;
  DynVector m_vNormal;
  DynPoint m_ptTexture;

  BOOL m_bBoundDynamic;
//////////////////////////////////////////////////////////////////////////////
// START: ASSIGNMENT 8: Handle dynamic point normals
//
  BOOL m_bAutoNormal;
  BOOL m_bAutoNormalDynamic;

  UINT_32 m_uiFaceCountMax;
  UINT_32 m_uiFaceCount;
//
// END  : ASSIGNMENT 8: Handle dynamic point normals
//////////////////////////////////////////////////////////////////////////////
};

//////////////////////////////////////////////////////////////////////
// Inlined Functions
//

inline VOID CSLIDEPoint::SetSurfaceID(CHAR *pcSurfaceID)
{
  m_pcSurfaceID = pcSurfaceID;
}

//////////////////////////////////////////////////////////////////////////////
// START: ASSIGNMENT 8: Handle dynamic point normals
//
inline VOID CSLIDEPoint::SetNormal(DynVector *pdynvNormal)
{
  SetAutoNormal(FALSE);
  m_vNormal.SetAll(pdynvNormal);
}
//
// END  : ASSIGNMENT 8: Handle dynamic point normals
//////////////////////////////////////////////////////////////////////////////

inline VOID CSLIDEPoint::SetTexture(DynPoint *pdynptTexture)
{
  m_ptTexture.SetAll(pdynptTexture);
}

inline CSLIDESurface *CSLIDEPoint::GetSurface()
{
  return m_pCSLIDESurface;
}

inline const Point &CSLIDEPoint::GetPoint()
{
  return m_pt;
}

inline const Vector &CSLIDEPoint::GetNormal()
{
  return m_vNormal;
}

inline const Point &CSLIDEPoint::GetTexture()
{
  return m_ptTexture;
}

inline BOOL CSLIDEPoint::GetBoundDynamic()
{
  return m_bBoundDynamic;
}

inline VOID CSLIDEPoint::SetBoundDynamic(BOOL bBoundDynamic)
{
  m_bBoundDynamic = bBoundDynamic;
}

inline BOOL CSLIDEPoint::GetAutoNormal()
{
  return m_bAutoNormal;
}

inline VOID CSLIDEPoint::SetAutoNormal(BOOL bAutoNormal)
{
  m_bAutoNormal = bAutoNormal;
}

inline BOOL CSLIDEPoint::GetAutoNormalDynamic()
{
  return m_bAutoNormalDynamic;
}

inline VOID CSLIDEPoint::SetAutoNormalDynamic(BOOL bAutoNormalDynamic)
{
  m_bAutoNormalDynamic = bAutoNormalDynamic;
  m_bDynamic = m_bDynamic || bAutoNormalDynamic;
}

#endif // __SLIDEPOINT_H__


