#include "slidesurface.h"

CSLIDESurface *CSLIDESurface::m_pCSLIDESurfaceDefault = NULL;

// Default Values
static const VECTOR4D cvColor = { 0.5, 0.5, 0.5, 0.0 };
static const VECTOR4D cvReflectivity = { 1.0, 1.0, 0.0, 0.0 };
static const FLOAT cfExponent = 1.0;
static const FLOAT cfMetallic = 0.0;

//////////////////////////////////////////////////////////////////////
// CSLIDESurface Class
//

CSLIDESurface::CSLIDESurface()
{
  ((Vector*)(&m_vColor))->Set(cvColor);
  m_vReflectivity.Set(cvReflectivity);
  m_dynfExponent.SetValue(cfExponent);
  m_dynfMetallic.SetValue(cfMetallic);
}

CSLIDESurface::~CSLIDESurface()
{
}

VOID CSLIDESurface::Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, 
			 CHAR *pcID)
{
  CSLIDEPrimitive::Init(pcFile, uiLine, uiColumn, pcID);

  m_vColor.Set(cvColor);
  m_vReflectivity.Set(cvReflectivity);
  m_dynfExponent.SetValue(cfExponent);
  m_dynfMetallic.SetValue(cfMetallic);
}

VOID CSLIDESurface::Uninit()
{
  CSLIDEPrimitive::Uninit();
}

BOOL CSLIDESurface::Initialized()
{
  return CSLIDEPrimitive::Initialized();
}

BOOL CSLIDESurface::CheckSemantics(CSLIDEWorld *pCSLIDEWorld)
{
  return CSLIDEPrimitive::CheckSemantics(pCSLIDEWorld);
}

BOOL CSLIDESurface::Valid()
{
  if ( !CSLIDEPrimitive::Valid() )
    {
      return FALSE;
    }

  return ( 0.0 <= m_vColor[R] && m_vColor[R] <= 1.0 &&
	   0.0 <= m_vColor[G] && m_vColor[G] <= 1.0 &&
	   0.0 <= m_vColor[B] && m_vColor[B] <= 1.0 &&
	   0.0 <= m_vReflectivity[0] && m_vReflectivity[0] <= 1.0 &&
	   0.0 <= m_vReflectivity[1] && m_vReflectivity[1] <= 1.0 &&
	   0.0 <= m_vReflectivity[2] && m_vReflectivity[2] <= 1.0 );
}

VOID CSLIDESurface::CheckDynamic()
{
  SetDynamic(m_vColor.GetDynamic() ||
	     m_vReflectivity.GetDynamic() ||
	     m_dynfExponent.GetDynamic() ||
	     m_dynfMetallic.GetDynamic());
}

VOID CSLIDESurface::UpdateState()
{
  m_vColor.Update();
  m_vReflectivity.Update();
  m_dynfExponent.Update();
  m_dynfMetallic.Update();
}

VOID CSLIDESurface::Push()
{
  if ( this != m_pCSLIDESurfaceDefault )
    {
      gilPushAttrib(GL_CURRENT_BIT);

      Render();
    }
}

VOID CSLIDESurface::Pop()
{
  if ( this != m_pCSLIDESurfaceDefault )
    {
      gilPopAttrib();
    }
}

VOID CSLIDESurface::Render()
{
  gilColor3fv(m_vColor.Get());
}

VOID CSLIDESurface::WriteSLIDE(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "%s %s\n",
	  gcppcSLIDEReservedWords[sliderwtSurface], m_pcID); 
  fprintf(fpOutputFile, "  %s ( %.3f %.3f %.3f )\n",
	  gcppcSLIDEReservedWords[sliderwtColor], 
	  m_vColor[R], m_vColor[G], m_vColor[B]); 
  fprintf(fpOutputFile, "  %s ( %.3f %.3f %.3f )\n",
	  gcppcSLIDEReservedWords[sliderwtReflectivity], 
	  m_vReflectivity[0], m_vReflectivity[1], m_vReflectivity[2]); 
  fprintf(fpOutputFile, "  %s %.3f\n",
	  gcppcSLIDEReservedWords[sliderwtExponent], 
	  m_dynfExponent.GetValue());
  fprintf(fpOutputFile, "  %s %.3f\n",
	  gcppcSLIDEReservedWords[sliderwtMetallic], 
	  m_dynfMetallic.GetValue());
  fprintf(fpOutputFile, "%s\n",
	  gcppcSLIDEReservedWords[sliderwtEndSurface]);
}
